{*******************************************************}
{                                                       }
{                     TIniClass                        }
{                                                       }
{           Copyright (c) 2000 by Serge Gusev           }
{                                                       }
{*******************************************************}

unit IniClass;

interface

uses Windows, Classes, IniFiles, Registry, Forms, SysUtils, TypInfo;

const
  SCommon = 'Common';

type
  { TIniClass }

  TSaveType = (stRegistry, stIniFile);

  TIniClass = class(TPersistent)
  private
    FSaveType: TSaveType;
    FFileName: string;
    procedure SetSaveType(const Value: TSaveType);
    function LoadFromRegister: Boolean;
    function LoadFromIniFile: Boolean;
    function SaveToRegister: Boolean;
    function SaveToIniFile: Boolean;
    procedure ReadObject(IniFile: TCustomIniFile; Section: string;
      AObject: TObject);
    procedure WriteObject(IniFile: TCustomIniFile; Section: string;
      AObject: TObject);
    procedure ReadStrings(IniFile: TCustomIniFile; Section: string;
      Strings: TStrings);
    procedure WriteStrings(IniFile: TCustomIniFile; Section: string;
      Strings: TStrings);
  protected
    function GetFileName: string; virtual;
    function LoadProperties(IniFile: TCustomIniFile): Boolean; virtual;
    function SaveProperties(IniFile: TCustomIniFile): Boolean; virtual;
    function IsStoredObject(AObject: TObject): Boolean; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    function Load: Boolean;
    function Save: Boolean;
    property SaveType: TSaveType read FSaveType write SetSaveType;
    property FileName: string read FFileName write FFileName;
  end;

implementation

var
  IniClasses: TList = nil;

{ TIniClass }

constructor TIniClass.Create;
begin
  inherited Create;
  FSaveType := stRegistry;
  if IniClasses = nil then IniClasses := TList.Create;
  IniClasses.Add(Self);
  FFileName := GetFileName
end;

destructor TIniClass.Destroy;
begin
  IniClasses.Remove(Self);
  if IniClasses.Count = 0 then
  begin
    IniClasses.Free;
    IniClasses := nil
  end;
  inherited Destroy
end;

procedure TIniClass.SetSaveType(const Value: TSaveType);
begin
  if Value <> FSaveType then
  begin
    FSaveType := Value;
    FFileName := GetFileName
  end
end;

function TIniClass.LoadFromRegister: Boolean;
var
  RegIniFile: TRegistryIniFile;
begin
  RegIniFile := TRegistryIniFile.Create(FFileName);
  try
    Result := LoadProperties(RegIniFile)
  finally
    RegIniFile.Free
  end
end;

function TIniClass.LoadFromIniFile: Boolean;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(FFileName);
  try
    Result := LoadProperties(IniFile)
  finally
    IniFile.Free
  end
end;

function TIniClass.SaveToRegister: Boolean;
var
  RegIniFile: TRegistryIniFile;
begin
  RegIniFile := TRegistryIniFile.Create(FFileName);
  try
    Result := SaveProperties(RegIniFile)
  finally
    RegIniFile.Free
  end
end;

function TIniClass.SaveToIniFile: Boolean;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(FFileName);
  try
    Result := SaveProperties(IniFile)
  finally
    IniFile.Free
  end
end;

procedure TIniClass.ReadObject(IniFile: TCustomIniFile; Section: string;
  AObject: TObject);
var
  TypeData: PTypeData;
  PropInfos: PPropList;
  I: Integer;
  PropValue: string;
  P: Longint;
begin
  if (AObject = nil) or (AObject.ClassInfo = nil) then Exit;
  TypeData := GetTypeData(AObject.ClassInfo);
  if (TypeData = nil) or (TypeData^.PropCount = 0) then Exit;
  with TypeData^ do
  begin
    GetMem(PropInfos, PropCount * SizeOf(Pointer));
    try
      GetPropInfos(AObject.ClassInfo, PropInfos);
      for I := 0 to PropCount - 1 do
        if PropInfos^[I].PropType^.Kind <> tkClass then
          if IniFile.ValueExists(Section, PropInfos^[I].Name) then
          begin
            PropValue := IniFile.ReadString(Section,
              PropInfos^[I].Name, PropValue);
            if PropInfos^[I].PropType^.Kind = tkChar then
              PropValue := IntToStr(Ord(PropValue[1]));
            SetPropValue(AObject, PropInfos^[I].Name, PropValue)
          end;
      for I := 0 to PropCount - 1 do
        if PropInfos^[I].PropType^.Kind = tkClass then
        begin
          P := GetPropValue(AObject, PropInfos^[I].Name);
          if P <> 0 then
            if TObject(P) is TStrings then
              ReadStrings(IniFile, PropInfos^[I].Name, TStrings(P))
            else
              ReadObject(IniFile, PropInfos^[I].Name, TObject(P))
        end
    finally
      FreeMem(PropInfos)
    end
  end
end;

function IsStoredProp(Instance: TObject; PropInfo: PPropInfo): Boolean;

   function IsDefaultValue: Boolean;
   begin
     if PropInfo^.PropType^.Kind in
       [tkInteger, tkChar, tkEnumeration, tkSet] then
       Result := GetOrdProp(Instance, PropInfo) = PropInfo^.Default
     else
       Result := GetPropValue(Instance, PropInfo^.Name) = Null
   end;

begin
  Result := TypInfo.IsStoredProp(Instance, PropInfo) and not IsDefaultValue
end;

procedure TIniClass.WriteObject(IniFile: TCustomIniFile; Section: string;
  AObject: TObject);
var
  TypeData: PTypeData;
  PropInfos: PPropList;
  I: Integer;
  PropValue: string;
  P: Longint;
begin
  with IniFile do
    if SectionExists(Section) then EraseSection(Section);
  if (AObject = nil) or (AObject.ClassInfo = nil)
    or not IsStoredObject(AObject) then Exit;
  TypeData := GetTypeData(AObject.ClassInfo);
  if (TypeData = nil) or (TypeData^.PropCount = 0) then Exit;
  with TypeData^ do
  begin
    GetMem(PropInfos, PropCount * SizeOf(Pointer));
    try
      GetPropInfos(AObject.ClassInfo, PropInfos);
      for I := 0 to PropCount - 1 do
        if PropInfos^[I].PropType^.Kind <> tkClass then
          if IsStoredProp(AObject, PropInfos^[I]) then
          begin
            PropValue := GetPropValue(AObject, PropInfos^[I].Name);
            if PropInfos^[I].PropType^.Kind = tkChar then
              PropValue := Char(StrToInt(PropValue));
            IniFile.WriteString(Section, PropInfos^[I].Name, PropValue)
          end;
      for I := 0 to PropCount - 1 do
        if PropInfos^[I].PropType^.Kind = tkClass then
        begin
          P := GetPropValue(AObject, PropInfos^[I].Name);
          if P <> 0 then
            if TObject(P) is TStrings then
              WriteStrings(IniFile, PropInfos^[I].Name, TStrings(P))
            else
              WriteObject(IniFile, PropInfos^[I].Name, TObject(P))
        end
    finally
      FreeMem(PropInfos)
    end
  end
end;

procedure TIniClass.ReadStrings(IniFile: TCustomIniFile; Section: string;
  Strings: TStrings);
begin
  with IniFile do
    if SectionExists(Section) then ReadSection(Section, Strings)
end;

procedure TIniClass.WriteStrings(IniFile: TCustomIniFile; Section: string;
  Strings: TStrings);
var
  I: Integer;
begin
  with IniFile, Strings do
  begin
    if SectionExists(Section) then EraseSection(Section);
    for I := 0 to Count - 1 do WriteString(Section, Strings[I], '')
  end
end;

function TIniClass.GetFileName: string;

  function GetRegistryName: string;
  var
    VerSize: Integer;
    Handle: DWord;
    P, Q: Pointer;
    Len: UINT;
    ValueName: string;
  begin
    Result := ChangeFileExt(ExtractFileName(Application.ExeName), '');
    { Thank to the RX library developers }
    VerSize := GetFileVersionInfoSize(PChar(Application.ExeName), Handle);
    if VerSize > 0 then
    begin
      GetMem(P, VerSize);
      try
        if GetFileVersionInfo(PChar(Application.ExeName), Handle, VerSize, P)
          and VerQueryValue(P, '\VarFileInfo\Translation', Q, Len)
          and (Q <> nil) then
          begin
            ValueName := '\StringFileInfo\'
              + IntToHex(MakeLong(HiWord(Longint(Q^)), LoWord(Longint(Q^))), 8)
              + '\CompanyName';
            if VerQueryValue(P, PChar(ValueName), Q, Len)
              and (Q <> nil) and (StrPas(Q) <> '') then
              Result := StrPas(Q) + '\' + Result
          end
      finally
        FreeMem(P)
      end;
      Result := 'Software\' + Result
    end
  end;

  function GetIniFileName: string;
  begin
    Result := ChangeFileExt(Application.ExeName, '.ini')
  end;

begin
  case FSaveType of
    stRegistry: Result := GetRegistryName;
    stIniFile: Result := GetIniFileName
    else
      Result := ''
  end
end;

function TIniClass.LoadProperties(IniFile: TCustomIniFile): Boolean;
begin
  Result := True;
  try
    ReadObject(IniFile, SCommon, Self)
  except
    Result := False
  end
end;

function TIniClass.SaveProperties(IniFile: TCustomIniFile): Boolean;
begin
  Result := True;
  try
    WriteObject(IniFile, SCommon, Self)
  except
    Result := False
  end
end;

function TIniClass.IsStoredObject(AObject: TObject): Boolean;
var
  TypeData: PTypeData;
  PropInfos: PPropList;
  I: Integer;
  P: Longint;
begin
  Result := False;
  if AObject.ClassInfo <> nil then
  begin
    TypeData := GetTypeData(AObject.ClassInfo);
    if (TypeData = nil) or (TypeData^.PropCount = 0) then Exit;
    with TypeData^ do
    begin
      GetMem(PropInfos, PropCount * SizeOf(Pointer));
      try
        GetPropInfos(AObject.ClassInfo, PropInfos);
        for I := 0 to PropCount - 1 do
        begin
          if PropInfos^[I].PropType^.Kind = tkClass then
          begin
            P := GetPropValue(AObject, PropInfos^[I].Name);
            Result := IsStoredObject(TObject(P))
          end
          else Result := IsStoredProp(AObject, PropInfos^[I]);
          if Result then Break
        end
      finally
        FreeMem(PropInfos)
      end
    end
  end
  else Result := AObject is TStrings
end;

function TIniClass.Load: Boolean;
begin
  case FSaveType of
    stRegistry: Result := LoadFromRegister;
    stIniFile: Result := LoadFromIniFile
    else
      Result := False
  end
end;

function TIniClass.Save: Boolean;
begin
  case FSaveType of
    stRegistry: Result := SaveToRegister;
    stIniFile: Result := SaveToIniFile
    else
      Result := False
  end
end;

initialization

finalization

  while IniClasses <> nil do TIniClass(IniClasses[0]).Free

end.
