{*******************************************************}
{                                                       }
{              Memory Inspector project                 }
{                                                       }
{              Memory Inspector object                  }
{                                                       }
{           Copyright (c) 2000 by Serge Gusev           }
{                                                       }
{*******************************************************}

unit MemInsp;

interface

uses Windows, Classes, GblClass, IndexLst;

type

  { TMemObjectInfo }

  PMemObjectInfo = ^TMemObjectInfo;
  TMemObjectInfo = record
    Address: Pointer;
    CallStack: TList;
    MemSize: Cardinal;
    MemObjectClass: TClass;
  end;  

function GetMemObjectCount: Integer;

function GetMemObjectInfo(Index: Integer): TMemObjectInfo;

function GetTotalSize: Integer;

function LockManager: Integer;

function UnlockManager: Integer;

var                                           
  CallStackCount: Integer = 4;

type

   TAllocateProc = procedure(Address: Pointer; AClass: TClass;
     var IsStore: Boolean);
   TDeallocateProc = procedure(Address: Pointer; AClass: TClass);
   TFreeManagerProc = procedure;

var
  AllocateProc: TAllocateProc = nil;
  DeallocateProc: TDeallocateProc = nil;
  FreeManagerProc: TFreeManagerproc = nil;

implementation

uses SysUtils, Graphics, Forms;

var
  OldManager: TMemoryManager;
  CriticalLock: TRTLCriticalSection;
  LockCount: Integer;

type

  { TMemObjectsList }

  TMemObjectsList = class(TIndexList)
  private
    function GetMemObjectInfo(Index: Integer): TMemObjectInfo;
  protected
    function CreateData: Pointer; override;
    procedure FreeData(AData: Pointer); override;
    procedure AssignData(Source, Dest: Pointer); override;
  public
    property MemObjects[Index: Integer]: TMemObjectInfo read GetMemObjectInfo; default;
  end;

{ TMemObjectsList }

function TMemObjectsList.GetMemObjectInfo(Index: Integer): TMemObjectInfo;
begin
  Result := PMemObjectInfo(Data[Index])^
end;

function TMemObjectsList.CreateData: Pointer;
var
  P: PMemObjectInfo;
begin
  New(P);
  with P^ do
  begin
    Address := nil;
    CallStack := TList.Create;
    CallStack.Capacity := CallStackCount;
    MemSize := 0;
    MemObjectClass := nil
  end;
  Result := P
end;

procedure TMemObjectsList.FreeData(AData: Pointer);
begin
  PMemObjectInfo(AData)^.CallStack.Free;
  FreeMem(AData)
end;

procedure TMemObjectsList.AssignData(Source, Dest: Pointer);
var
  S, D: PMemObjectInfo;
  I: Integer;
begin
  S := PMemObjectInfo(Source);
  D := PMemObjectInfo(Dest);
  D^.Address := S^.Address;
  D^.CallStack.Count := S.CallStack.Count;
  for I := 0 to S^.CallStack.Count - 1 do
    D^.CallStack[I] := S^.CallStack[I];
  D^.MemSize := S^.MemSize;
  D^.MemObjectClass := S^.MemObjectClass  
end;

type
  PPointer = ^Pointer;

var
  MemObjectsList: TMemObjectsList;

  StrGetMem: Pointer = nil;
  StrFreeMem: Pointer = nil;
  StrReallocMem: Pointer = nil;

procedure GetCallStack(AClass: TClass; CallStack: TList);

  function GetStackAddress: Integer;
  asm
          MOV  EAX, ESP
  end;

var
  Address: PPointer;
  MemInfo: TMemoryBasicInformation;
begin
  LockManager;
  try
    if AClass = nil then
      Address := Pointer(GetStackAddress + 112)
    else
      Address := Pointer(GetStackAddress + 120);
    while CallStack.Count < CallStackCount do
    begin
      VirtualQuery(Address, MemInfo, SizeOf(MemInfo));
      if MemInfo.Protect = PAGE_NOACCESS then Break
      else
      begin
        if IsExecuting(Address^) then
          CallStack.Add(Pointer(Address^));
        Inc(Integer(Address), 4)
      end
    end
  finally
    UnlockManager
  end
end;

function GetMem(Size: Integer): Pointer;

  function GetCallAddress: Pointer;
  asm
          MOV   EAX, [ESP + 60]
  end;

  function GetClass: TClass;
  asm
          MOV   EAX, [ESP + 60]
          CALL  IsClass
          TEST  EAX, EAX
          JZ    @Exit
          MOV   EAX, [ESP + 60]
    @Exit:
  end;

var
  AClass: TClass;
  IsStore: Boolean;
  MemObjectInfo: PMemObjectInfo;
begin
  EnterCriticalSection(CriticalLock);
  try
    Result := OldManager.GetMem(Size);
    if LockCount = 0 then
    begin
      LockManager;
      try
        IsStore := GetCallAddress <> StrGetMem;
        if IsStore then AClass := GetClass else AClass := nil;
        if IsStore and Assigned(AllocateProc) then
          AllocateProc(Result, AClass, IsStore);
        if IsStore then
        begin
          MemObjectInfo := MemObjectsList.Add(Integer(Result));
          if MemObjectInfo <> nil then
            with MemObjectInfo^ do
            begin
              Address := Result;
              if CallStackCount <> 0 then
                GetCallStack(AClass, CallStack);
              MemSize := Size;
              MemObjectClass := AClass
            end
        end
      finally
        UnlockManager
      end
    end
  finally
    LeaveCriticalSection(CriticalLock)
  end
end;

function FreeMem(P: Pointer): Integer;

  function GetCallAddress: Pointer;
  asm
          MOV   EAX, [ESP + 40]
  end;

var
  AClass: TClass;
begin
  asm
          PUSH   EAX
          MOV    EAX, [EAX]
          MOV    AClass, EAX
          POP    EAX
  end;
  if AClass = TMultiReadExclusiveWriteSynchronizer then
  begin
    if Assigned(FreeManagerProc) then FreeManagerProc;
    SetMemoryManager(OldManager);
    MemObjectsList.Free;
    MemObjectsList := nil;
    DeleteCriticalSection(CriticalLock);
    Result := 0
  end
  else
  begin
    AClass := nil;
    Result := OldManager.FreeMem(P);
    if (LockCount = 0) and (GetCallAddress <> StrFreeMem) then
    begin
      LockManager;
      EnterCriticalSection(CriticalLock);
      try
        if Assigned(DeallocateProc) then DeallocateProc(P, AClass);
        MemObjectsList.Remove(Integer(P))
      finally
        LeaveCriticalSection(CriticalLock);
        UnlockManager
      end
    end
  end
end;

function ReallocMem(P: Pointer; Size: Integer): Pointer;

  function GetCallAddress: Pointer;
  asm
          MOV   EAX, [ESP + 76]
  end;

var
  Index: Integer;
  AClass: TClass;
  MemObjectInfo: PMemObjectInfo;
begin
  Index := -1;
  AClass := nil;
  EnterCriticalSection(CriticalLock);
  try
    if LockCount = 0 then
    begin
      LockManager;
      try
        if GetCallAddress = StrReallocMem then Index := -1
        else
        begin
          Index := MemObjectsList.IndexOf(Integer(P));
          if Index <> -1 then
            AClass := MemObjectsList.MemObjects[Index].MemObjectClass
        end
      finally
        UnlockManager
      end
    end;
    Result := OldManager.ReallocMem(P, Size);
    if (LockCount = 0) and (Index <> -1) then
    begin
      LockManager;
      try
        if Result <> P then MemObjectsList.Delete(Index);
        MemObjectInfo := MemObjectsList.Add(Integer(Result));
        if MemObjectInfo <> nil then
          with MemObjectInfo^ do
          begin
            Address := Result;
            if CallStackCount <> 0 then
              GetCallStack(AClass, CallStack);
            MemSize := Size;
            MemObjectClass := AClass
          end
      finally
        UnlockManager
      end
    end
  finally
    LeaveCriticalSection(CriticalLock)
  end;
  if AClass = TMultiReadExclusiveWriteSynchronizer then
  begin
    SetMemoryManager(OldManager);
    MemObjectsList.Free;
    DeleteCriticalSection(CriticalLock)
  end
end;

var
  Manager: TMemoryManager = (GetMem: GetMem; FreeMem: FreeMem;
    ReallocMem: ReallocMem);

function CheckGetMem(Size: Integer): Pointer;

  function GetCallAddress: Pointer;
  asm
          MOV   EAX, [ESP + 28]
  end;

begin
  try
    StrGetMem := GetCallAddress
  finally
    Result := OldManager.GetMem(Size)
  end
end;

function CheckFreeMem(P: Pointer): Integer;

   function GetCallAddress: Pointer;
   asm
          MOV   EAX, [ESP + 28]
   end;

begin
  try
    StrFreeMem := GetCallAddress
  finally
    Result := OldManager.FreeMem(P)
  end  
end;

function CheckReallocMem(P: Pointer; Size: Integer): Pointer;

  function GetCallAddress: Pointer;
  asm
          MOV   EAX, [ESP + 36]
  end;

begin
  try
    StrReallocMem := GetCallAddress
  finally
    Result := OldManager.ReallocMem(P, Size)
  end
end;

var
  CheckManager: TMemoryManager = (GetMem: CheckGetMem;
    FreeMem: CheckFreeMem; ReallocMem: CheckReallocMem);

  CheckStr: string;

procedure CheckStrAddresses;
begin
  SetMemoryManager(CheckManager);
  try
    CheckStr := 'Check';
    SetLength(CheckStr, 25) ;
    CheckStr := ''
  except
    SetMemoryManager(OldManager)
  end
end;

function GetMemObjectCount: Integer;
begin
  if MemObjectsList = nil then Result := 0
  else
    Result := MemObjectsList.Count
end;

function GetMemObjectInfo(Index: Integer): TMemObjectInfo;
begin
  if MemObjectsList = nil then
    with Result do
    begin
      Address := nil;
      CallStack := nil;
      MemSize := 0;
      MemObjectClass := nil
    end
  else Result := MemObjectsList[Index]
end;

function GetTotalSize: Integer;
var
  I: Integer;
begin
  Result := 0;
  if MemObjectsList <> nil then
    for I := 0 to MemObjectsList.Count - 1 do
      Inc(Result, MemObjectsList.MemObjects[I].MemSize)
end;

function LockManager: Integer;
begin
  Inc(LockCount);
  Result := LockCount
end;

function UnlockManager: Integer;
begin
  if LockCount > 0 then Dec(LockCount);
  Result := LockCount
end;

initialization

  MemObjectsList := TMemObjectsList.Create;
  InitializeCriticalSection(CriticalLock);
  GetMemoryManager(OldManager);
  CheckStrAddresses;
  SetMemoryManager(Manager)

end.
